from timeit import timeit
import pygame_ecs
import random
from sys import argv

try:
    arg = argv[1]
    if arg != "perfect":
        arg = "imperfect"
except IndexError:
    arg = "perfect"

WIDTH = 800
HEIGHT = 800
ENTITY_AMOUNT = 1_000 * 4

# TODO: put entities as a list in entity_manager
# Also make the component manager be kept as a reference in the system_manager
# I can insert EntityManager reference into BaseSystem Subclasses using the same method I've used with components nvm i cant lol
# actually add the systems into


class Position(pygame_ecs.BaseComponent):
    def __init__(self, x: int, y: int):
        super().__init__()
        self.x = x
        self.y = y


class Velocity(pygame_ecs.BaseComponent):
    def __init__(self, vec: list[int | float]) -> None:
        super().__init__()
        self.vec = vec


class BallPhysics(pygame_ecs.BaseSystem):
    def __init__(self) -> None:
        super().__init__(required_component_types=[Position, Velocity])

    def update_entity(self, entity, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        vel: Velocity = entity_components[Velocity]  # type: ignore
        pos.x += vel.vec[0]  # type: ignore
        pos.y += vel.vec[1]  # type: ignore
        if pos.x > WIDTH or pos.x < 0:
            vel.vec[0] *= -1
        if pos.y > HEIGHT or pos.y < 0:
            vel.vec[1] *= -1


component_manager = pygame_ecs.ComponentManager()
entity_manager = pygame_ecs.EntityManager(component_manager)
system_manager = pygame_ecs.SystemManager(entity_manager, component_manager)
ball_physics = BallPhysics()
system_manager.add_system(ball_physics)
component_manager.init_components()

for _ in range(ENTITY_AMOUNT):
    center = (
        random.randint(0, WIDTH),
        random.randint(0, HEIGHT),
    )
    radius = random.randint(2, 12)
    color = [random.randint(0, 255) for _ in range(3)]
    vel = [
        (random.random() - 0.5) * 400 / 1000,
        (random.random() - 0.5) * 400 / 1000,
    ]
    entity = entity_manager.add_entity()
    component_manager.add_component(entity, Position(center[0], center[1]))
    if arg == "perfect":
        component_manager.add_component(entity, Velocity(vel))

for entity in entity_manager.entities.keys():
    entity_manager.kill_entity(entity)
entity_manager._clear_limbo()

for _ in range(
    ENTITY_AMOUNT + 1_000
):  # ensure that killing and then spawning entities doesnt break anything
    center = (
        random.randint(0, WIDTH),
        random.randint(0, HEIGHT),
    )
    radius = random.randint(2, 12)
    color = [random.randint(0, 255) for _ in range(3)]
    vel = [
        (random.random() - 0.5) * 400 / 1000,
        (random.random() - 0.5) * 400 / 1000,
    ]
    entity = entity_manager.add_entity()
    component_manager.add_component(entity, Position(center[0], center[1]))
    if arg == "perfect":
        component_manager.add_component(entity, Velocity(vel))

REPEAT = 1_000

res = timeit(lambda: system_manager.update_entities(), number=REPEAT)  # type: ignore
print(
    f"Took {res/REPEAT} roughly for each frame, using {len(entity_manager.entities)} entities, setting: {arg}"
)



import typing
from pygame_ecs.components.base_component import BaseComponent
from pygame_ecs.managers.component_manager import ComponentInstanceType
from pygame_ecs.managers.entity_manager import EntityManager
from pygame_ecs.entity import Entity


class BaseSystem:
    # typing.Type specifies that it will take subclasses of this class
    def __init__(
        self, required_component_types: list[typing.Type[BaseComponent]]
    ) -> None:
        self.required_component_types = required_component_types
        self.entity_manager: EntityManager

    def update_entity(
        self,
        entity: Entity,
        entity_components: dict[typing.Type[BaseComponent], ComponentInstanceType],
    ):
        """This function is called for every entity that has the required components if the required component size is bigger than 0.

        Args:
            entity (Entity): Entity instance

            entity_components (dict[typing.Type[BaseComponent], ComponentInstanceType]): Components of the entity
        """
        pass

    def update(self):
        """This function is called only once a frame if the required components is 0."""
        pass

    def __str__(self) -> str:
        return f"<{type(self).__name__}>"



import typing

from pygame_ecs.systems.base_system import BaseSystem
from pygame_ecs.managers.component_manager import ComponentManager
from pygame_ecs.managers.entity_manager import EntityManager

SystemType = typing.TypeVar("SystemType", bound=BaseSystem)


class SystemManager:
    def __init__(
        self, entity_manager: EntityManager, component_manager: ComponentManager
    ) -> None:
        self.entity_manager = entity_manager
        self.component_manager = component_manager
        self.systems: list[BaseSystem] = []

    def add_system(self, system: SystemType):
        system.entity_manager = self.entity_manager
        self.systems.append(system)

    def remove_system(self, system: SystemType):
        self.systems.remove(system)

    def update_entities(self):
        """Updates all of the systems that are active.
        NOTE: For updating values of systems, just set their values before calling this function.
        """
        for system in self.systems:
            if len(system.required_component_types) > 0:
                for entity in self.entity_manager.entities.keys():
                    has_components = True
                    components_to_give = {}
                    for comp_type in system.required_component_types:
                        try:
                            comp = self.component_manager.components[comp_type][entity]
                            components_to_give[type(comp)] = comp
                        except KeyError:
                            self.component_manager.components[comp_type]
                            has_components = False
                            break
                    if has_components:
                        system.update_entity(entity, components_to_give)
            else:
                system.update()
        self.entity_manager._clear_limbo()



import typing
import pygame
import random
import pygame_ecs
from pygame._sdl2 import Renderer, Texture, Window, get_drivers

from pygame_ecs.components.base_component import BaseComponent

pygame.init()

ENTITY_AMOUNT = 1
WIDTH = 800
HEIGHT = 600


class Position(pygame_ecs.BaseComponent):
    __slots__ = ("x", "y")

    def __init__(self, x: int, y: int):
        super().__init__()
        self.x = x
        self.y = y


class BallRenderer(pygame_ecs.BaseComponent):
    __slots__ = ("radius", "color")

    def __init__(self, radius: int, color) -> None:
        super().__init__()
        self.radius = radius
        self.color = color


class Velocity(pygame_ecs.BaseComponent):
    __slots__ = "vec"

    def __init__(self, vec: pygame.math.Vector2) -> None:
        super().__init__()
        self.vec = vec


class BallDrawSystem(pygame_ecs.BaseSystem):
    def __init__(self, texture: Texture) -> None:
        super().__init__(required_component_types=[Position, BallRenderer])
        self.texture = texture

    def update_entity(self, entity, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        ball_renderer: BallRenderer = entity_components[BallRenderer]  # type: ignore

        self.texture.color = ball_renderer.color  # type: ignore
        self.texture.draw(
            None, (pos.x, pos.y, ball_renderer.radius, ball_renderer.radius)
        )


class BallPhysics(pygame_ecs.BaseSystem):
    def __init__(self, screen) -> None:
        super().__init__(required_component_types=[Position, Velocity])
        self.dt = 0
        self.screen = screen

    def update_entity(self, entity, entity_components):
        pos: Position = entity_components[Position]  # type: ignore
        vel: Velocity = entity_components[Velocity]  # type: ignore
        pos.x += vel.vec.x * self.dt  # type: ignore
        pos.y += vel.vec.y * self.dt  # type: ignore
        if pos.x > WIDTH or pos.x < 0:
            vel.vec.x *= -1
        if pos.y > HEIGHT or pos.y < 0:
            vel.vec.y *= -1


def add_entity(component_manager, entity_manager, amount=1):
    for _ in range(amount):
        center = (
            random.randint(0, WIDTH),
            random.randint(0, HEIGHT),
        )
        radius = random.randint(5, 15)
        color = [random.randint(0, 255) for _ in range(3)]
        color.append(255)
        vel = pygame.math.Vector2(
            (random.random() - 0.5) * 400 / 1000,
            (random.random() - 0.5) * 400 / 1000,
        )
        entity = entity_manager.add_entity()
        component_manager.add_component(entity, Position(center[0], center[1]))
        component_manager.add_component(entity, Velocity(vel))
        component_manager.add_component(entity, BallRenderer(radius, color))


def remove_entity(entity_manager: pygame_ecs.EntityManager, amount=1):
    if len(entity_manager.entities) < amount:
        return
    for _ in range(amount):
        entity = random.choice(list(entity_manager.entities.keys()))
        try:
            entity_manager.kill_entity(entity)
        except pygame_ecs.EntityAlreadyInLimbo:
            pass


class BallAdd(pygame_ecs.BaseSystem):
    def __init__(self, component_manager, entity_manager) -> None:
        super().__init__(required_component_types=[])
        self.component_manager = component_manager
        self.entity_manager = entity_manager

    def update(self):
        if pygame.mouse.get_pressed()[0]:
            add_entity(component_manager, entity_manager, amount=50)
        elif pygame.mouse.get_pressed()[2]:
            remove_entity(entity_manager, amount=50)


screen = pygame.display.set_mode((WIDTH, HEIGHT))
window = Window.from_display_module()


print(list(get_drivers()))
# change based on drivers
renderer = Renderer(window, index=3, accelerated=1)


texture = Texture.from_surface(renderer, pygame.image.load("test/circle.png"))

component_manager = pygame_ecs.ComponentManager()
entity_manager = pygame_ecs.EntityManager(component_manager)
system_manager = pygame_ecs.SystemManager(entity_manager, component_manager)
ball_physics = BallPhysics(screen)
ball_draw_system = BallDrawSystem(texture=texture)
ball_add = BallAdd(component_manager, entity_manager)
system_manager.add_system(ball_draw_system)
system_manager.add_system(ball_physics)
system_manager.add_system(ball_add)
component_manager.init_components()

add_entity(component_manager, entity_manager, amount=ENTITY_AMOUNT)


clock = pygame.time.Clock()
dt = 0

renderer.draw_color = (
    0,
    0,
    0,
    255,
)  # type: ignore
# renderer.draw_color is used for clearing the screen and drawing primitives

running = True
while running:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False

        if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            running = False
    pygame.display.set_caption(f"FPS: {clock.get_fps()} COUNT: {len(entity_manager.entities)}")
    ball_physics.dt = dt
    renderer.clear()
    system_manager.update_entities()
    renderer.present()
    dt = clock.tick(60)

pygame.quit()



from pygame_ecs.entity import Entity
from pygame_ecs.managers.component_manager import ComponentManager
from pygame_ecs.exceptions import EntityAlreadyInLimbo


class EntityManager:
    def __init__(self, component_manager: ComponentManager) -> None:
        self.component_manager = component_manager
        self.entities: dict[Entity, None] = {}
        self.dead_entities: list[Entity] = []
        self._limbo: dict[Entity, None] = {}
        self.count: int = 0

    def add_entity(self):
        if len(self.dead_entities) > 0:
            entity = self.dead_entities.pop()
        else:
            entity = Entity(self.count)
            self.count += 1
        self.entities[entity] = None
        return entity

    def kill_entity(self, entity: Entity):
        """Kills an entity by moving it into `limbo`. Entities in limbo is removed from  entities once ._clear_limbo is called.
        NOTE: ._clear_limbo is called automatically when all of the active systems are done updating.

        Args:
            entity (Entity): an Entity object

        Raises:
            EntityAlreadyInLimbo: An entity is already in limbo(already has been killed)
        """
        try:
            self._limbo[entity]
            raise EntityAlreadyInLimbo(entity)
        except KeyError:  # not in limbo
            self._limbo[entity] = None
        for component_type in self.component_manager.components.keys():
            try:
                del self.component_manager.components[component_type][entity]
            except KeyError:
                pass

    def _clear_limbo(self):
        """Function to get rid of entities that have been killed that frame.
        You probably won't need to call this yourself, it will be called automatically.
        """
        for entity, _ in self._limbo.items():
            self.dead_entities.append(entity)
            del self.entities[entity]
        self._limbo = {}
